/// PAQ8I file compressor/archiver.
//(C) 2006, Matt Mahoney under GPL, http://www.gnu.org/licenses/gpl.txt

#include <classes.hpp>

// Compress/decompress files.  Usage: PAQ8I archive files...
// If archive does not exist, it is created and the named files are
// compressed.  If there are no file name arguments after the archive,
// then file names are read from input up to a blank line or EOF.
// If archive already exists, then the files in the archive are either
// extracted, or compared if the file already exists.  The files
// are extracted to the names listed on the command line in the
// order they were stored, defaulting to the stored names.

int main(int argc, char** argv) {
  clock_t start_time=clock();  // start timer
  long total_size=0;  // uncompressed size of all files

    // Print help message
    if (argc<2) {
      printf(PROGNAME " archiver (C) 2006, Matt Mahoney.\n"
        "Free under GPL, http://www.gnu.org/licenses/gpl.txt\n\n"
        "To compress:\n"
		"  " PROGNAME " -level archive files... (archive will be created)\n"
        "or (Windows):\n"
		"  dir/b | " PROGNAME " archive  (file names read from input)\n"
		"\n"
        "level: -0 = store, -1 -2 -3 = faster (uses 21, 28, 42 MB)\n"
        "-4 -5 -6 -7 -8 = smaller (uses 120, 225, 450, 900, 1800 MB)\n"
        "level -%d is default\n"
        "\n"
        "To extract or compare:\n"
        "  " PROGNAME " archive  (extracts to stored names)\n"
        "\n"
        "To view contents: more < archive\n"
        "\n",
        DEFAULT_OPTION);
      return 0;
    }

    // Get option
  int option='4';
  if (argv[1][0]=='-')
    option=argv[1][1], ++argv, --argc;
  if (option<32) option=32;
  if (option<'0'||option>'9')
    fprintf(stderr, "Bad option -%c (use -0 to -9)\n", option), exit(1);

  // Test for archive.  If none, create one and write a header.
  // The first line is PROGNAME.  This is followed by a list of
  // file sizes (as decimal numbers) and names, separated by a tab
  // and ending with \r\n.  The last entry is followed by ^Z
  Mode mode=DECOMPRESS;
  FILE *f=fopen(argv[1], "rb");
  if (!f) {
    mode=COMPRESS;

    f=fopen(argv[1], "wb");
    if (!f)
      perror(argv[1]), exit(1);
    fprintf(f, "%s -%c\r\n", PROGNAME, option);

    // Get filenames
    if (argc==2)
      printf("Enter names of files to compress, followed by blank line\n");
    int i=2;
	std::multimap <int,std::string> filetypes;
	std::multimap <int,std::string>::iterator it;
	std::vector <std::string> filenames;
    char *filename;

    while (true) {

      if (argc==2) {
        filename=getline();
        if (!filename || !filename[0])
          break;
      }
      else {
        if (i==argc)
          break;
        filename=argv[i++];
      }
	  filenames.push_back(filename);
    } // end while

	// Detect file types and make map of it
	for (i=0; i<filenames.size(); i++) {
	  Filter::make(filenames[i].c_str(), NULL);
   	  std::pair<int,std::string> p((1<<31-1)-(filetype*8*16*2048+preprocFlag+(wrt.longDict+1)*16*2048+(wrt.shortDict+1)*2048),filenames[i]);
	  PRINT_DICT((" %s\n",filenames[i].c_str()));
	  filetypes.insert(p);
	} 

	// Write filesize, filename for each file in archive header.
	for (it=filetypes.begin(); it!=filetypes.end(); it++) // multimap is sorted by filetype
		store_in_header(f,(char*)it->second.c_str(),total_size);

    fputc(26, f);  // EOF
    fclose(f);
  }

  // Read existing archive. Two pointers (header and body) track the
  // current filename and current position in the compressed data.
  if (mode==COMPRESS)
    f=fopen(argv[1], "r+b");
  else
    f=fopen(argv[1], "rb");

  // Output file must already exist 
  if (!f)
    perror(argv[1]), exit(1);

  long header, body;  // file positions in header, body
  char *filename=getline(f);  // check header
  if (!filename || strncmp(filename, PROGNAME " -", strlen(PROGNAME)+2))
    fprintf(stderr, "%s: not a " PROGNAME " file\n", argv[1]), exit(1);
  option=filename[strlen(filename)-1];
  level=option-'0';
  if (level<0||level>9) level=DEFAULT_OPTION;
  buf.setsize(MEM*8);
  header=ftell(f);

  // Initialize encoder at end of header
  if (mode==COMPRESS)
    fseek(f, 0, SEEK_END);
  else {  // body starts after ^Z in file
    int c;
    while ((c=getc(f))!=EOF && c!=26)
      ;
    if (c!=26)
      fprintf(stderr, "Archive %s is incomplete\n", argv[1]), exit(1);
  }
  Encoder en(mode, f);
  body=ftell(f);

  // Compress/decompress files listed on command line, or header if absent.
  int filenum=1;  // command line index
  total_size=0;
  fseek(f, header, SEEK_SET);
  while ((filename=getline(f))!=0) {
    size=atol(filename);  // parse size and filename, separated by tab
    total_size+=size;
    while (*filename && *filename!='\t')
      ++filename;
    if (*filename=='\t')
      ++filename;
    printf("%ld\t%s", size, filename);
    fsize=(int)size;
 /*   if (mode==DECOMPRESS && ++filenum<argc  // doesn't work with sorting depend on type of file
        && strcmp(argv[filenum], filename)) {
      printf(" -> %s", argv[filenum]);
      filename=argv[filenum];
    } */
    printf(": ");
    if (size<0 || total_size<0) break;
    header=ftell(f);
    fseek(f, body, SEEK_SET);

    // If file exists in COMPRESS mode, compare, else compress/decompress
    FILE *fi=fopen(filename, "rb");
    if (mode==COMPRESS) {
      if (!fi) perror(filename), exit(1);
      filetype=0; // BMP image = 0, no preprocessing
      Filter* fp=Filter::make(filename, &en);   // sets filetype
      fp->compress(fi, size); // Actual compression
      printf(" -> %4ld   \n", ftell(f)-body); // Report on compressed size
      delete fp;
    }
    else {  // DECOMPRESS, first byte determines filter type
      filetype=0;
      filetype=en.decompress();
      Filter* fp=Filter::make(0, &en);
      if (fi)
        fp->compare(fi, size);
      else {  // extract
        fi=fopen(filename, "wb");
        if (fi)
          fp->decompress(fi, size);
        else {
          perror(filename);
          fp->skip(size);
        }
      }
      delete fp;
    }
    if (fi)
      fclose(fi);
    body=ftell(f);
    fseek(f, header, SEEK_SET);
  }
  fseek(f, body, SEEK_SET);
  en.flush();

  // Print stats
  if (f) {
    if (mode==DECOMPRESS && filenum<argc-1)
      printf("No more files to extract\n");
    fseek(f, 0, SEEK_END);
    long compressed_size=ftell(f);
    if (mode==COMPRESS)
      printf("%ld -> %ld", total_size, compressed_size);
    else
      printf("%ld -> %ld", compressed_size, total_size);
    start_time=clock()-start_time;
    if (compressed_size>0 && total_size>0 && start_time>0) {
      printf(" (%1.4f bpc) in %1.2f sec (%1.3f KB/sec)",
        8.0*compressed_size/total_size,
        (double)start_time/CLOCKS_PER_SEC,
        0.001*total_size*CLOCKS_PER_SEC/start_time);
    }
    printf("\n");
    fclose(f);
  }
  return 0;
}
